# The LearningOnline Network
# Bulletin Board Handler
#
# $Id: lonbulletin.pm,v 1.31 2005/10/30 02:39:09 www Exp $
#
# Copyright Michigan State University Board of Trustees
#
# This file is part of the LearningOnline Network with CAPA (LON-CAPA).
#
# LON-CAPA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LON-CAPA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LON-CAPA; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# /home/httpd/html/adm/gpl.txt
#
# http://www.lon-capa.org/
#

package Apache::lonbulletin;

use strict;
use Apache::Constants qw(:common);
use Apache::loncommon;
use Apache::lonnet;
use Apache::lontexconvert;
use Apache::lonfeedback;
use Apache::lonlocal;

sub handler {
    my $r = shift;
    &Apache::loncommon::content_type($r,'text/html');
    $r->send_http_header;
    return OK if $r->header_only;
    my $target=$env{'form.grade_target'};

# ------------------------------------------------------------ Print the screen
    if ($target ne 'tex') {
	my $html=&Apache::lonxml::xmlbegin();
	$r->print(<<ENDDOCUMENT);
$html
<head>
<title>The LearningOnline Network with CAPA</title>
ENDDOCUMENT
} else {
	$r->print(&Apache::lonprintout::print_latex_header($env{'form.latex_type'}));
    }
    my (undef,undef,undef,undef,$marker)=split(/\//,$r->uri);
# Is this even in a course?
    unless ($env{'request.course.id'}) {
	$r->print('</head><body>Not in a course</body></html>');
        return OK;
    }

    $marker=~s/\D//g;

    unless ($marker) {
	$r->print('<body>Invalid call</body>');
        return OK;
    }

    my $dom = $env{'course.'.$env{'request.course.id'}.'.domain'};
    my $crs = $env{'course.'.$env{'request.course.id'}.'.num'};

# --------------------------------------------------------- The syllabus fields
    my %syllabusfields=&Apache::lonlocal::texthash(
       'aaa_title'         => 'Topic',
       'bbb_content'       => 'Task',
       'ccc_webreferences' => 'Web References');

# ------------------------------------------------------------ Get Query String
    &Apache::loncommon::get_unprocessed_cgi
                ($ENV{'QUERY_STRING'},['forcestudent','forceedit','register','origpage']);
# ----------------------------------------------------- Force menu registration
    my $addentries='';
    if ($env{'form.register'}) {
       $addentries=' onLoad="'.&Apache::lonmenu::loadevents();
       if ($env{'form.origpage'}) {
           $addentries .= "document.location='#newpost';";
       }
       $addentries .= '" onUnload="'.&Apache::lonmenu::unloadevents().'"';
       $r->print(&Apache::lonmenu::registerurl(1));
    } elsif ($env{'form.origpage'}) {
         $addentries = "onLoad=\"document.location='#newpost';\"";
    }
# --------------------------------------------------------------- Force Student
    my $forcestudent='';
    if ($env{'form.forcestudent'}) { $forcestudent='student'; }

    my $forceedit='';
    if ($env{'form.forceedit'}) { $forceedit='edit'; }

    my %syllabus=&Apache::lonnet::dump('bulletinpage_'.$marker,$dom,$crs);
       
# --------------------------------------- There is such a user, get environment
    if ($target ne 'tex') {   
	$r->print('</head>'.&Apache::loncommon::bodytag
                  ("Bulletin Board/Discussion",$forcestudent,$addentries,'',$dom,
                  $env{'form.register'}));
    }
    my $allowed=&Apache::lonnet::allowed('mdc',$env{'request.course.id'});

    my $privileged=$allowed;
    if (($syllabus{'uploaded.lastmodified'}) && (!$forceedit)) {
 	$forcestudent='student';
    }
    if ($target ne 'tex') { $r->print('<table><tr><td>'); }
      if ($forcestudent or $target eq 'tex') { $allowed=0; }
 
      if ($allowed) {
          $r->print(
	  '<p>'.
&Apache::loncommon::help_open_topic('Uploaded_Templates_TextBoxes','Help with filling in text boxes').'<br /><a href="'.$r->uri.'?forcestudent=1"><font size="+1">'.&mt('Show Student View').'</font></a>'.
 &Apache::loncommon::help_open_topic('Uploaded_Templates_PublicView').'</p>');
      }  elsif ($privileged and $target ne 'tex') {
	  $r->print('<a href="'.$r->uri.'?forceedit=edit"><font size="+1">'
		    .&mt('Edit').'</font></a>');
      } 

      if (($env{'form.uploaddoc.filename'}) &&
          ($env{'form.storeupl'}) && ($allowed)) {
 	  if ($env{'form.uploaddoc.filename'}=~/\.(gif|jpg|png|jpeg)$/i) {
	      if ($syllabus{'uploaded.photourl'}) {
		  &Apache::lonnet::removeuploadedurl($syllabus{'uploaded.photourl'});
	      }
	      $syllabus{'uploaded.photourl'}=
		  &Apache::lonnet::userfileupload('uploaddoc',1,'bulletin');
 	  }
          $syllabus{'uploaded.lastmodified'}=time;
          &Apache::lonnet::put('bulletinpage_'.$marker,\%syllabus,$dom,$crs);
       }
       if (($allowed) && ($env{'form.storesyl'})) {
	   foreach (keys %syllabusfields) {
               my $field=$env{'form.'.$_};
               $field=~s/\s+$//s;
               $field=&Apache::lonfeedback::clear_out_html($field,1);
	       $syllabus{$_}=$field;
           }
           $syllabus{'uploaded.lastmodified'}=time;
           &Apache::lonnet::put('bulletinpage_'.$marker,\%syllabus,$dom,$crs);
       }

# ---------------------------------------------------------------- Get syllabus
    if (($syllabus{'uploaded.lastmodified'}) || ($allowed)) {
       if ($syllabus{'uploaded.photourl'}) {
	   &Apache::lonnet::allowuploaded('/adm/syllabus',
					  $syllabus{'uploaded.photourl'});
	   $r->print('<img src="'.$syllabus{'uploaded.photourl'}.
	       '" align="right" />');
       }
       if ($allowed) {
           $r->print(
	 '<form method="post" enctype="multipart/form-data">'.
 	 '<input type="hidden" name="forceedit" value="edit" />'.
         '<h3>'.&mt('Upload a Photo').'</h3>'.
         '<input type="file" name="uploaddoc" size="50">'.
         '<input type="submit" name="storeupl" value="'.&mt('Upload').'">'.
	 '</form><form method="post">');
       }
       foreach (sort keys %syllabusfields) {
          if (($syllabus{$_}) || ($allowed)) {
              my $message=$syllabus{$_};
	      &Apache::lonfeedback::newline_to_br(\$message);
	      $message
        =~s/(https*\:\/\/[^\s]+)/\<a href=\"$1\"\>\<tt\>$1\<\/tt\>\<\/a\>/g;
	      if ($allowed) {
		  $message=&Apache::lonspeller::markeduptext($message);
	      }
	      $message=&Apache::lontexconvert::msgtexconverted($message);
	      unless ($_ eq 'aaa_title') {
		if (($_ ne 'bbb_content') || ($allowed)) {
		    if ($target ne 'tex') {
			$r->print('<h3>'.$syllabusfields{$_}.'</h3>');
		    } else {
			$r->print('\\\\\textbf{'.$syllabusfields{$_}.'}\\\\');
		    }
		}
		if ($target ne 'tex') {
		    $r->print('<blockquote>'.
                              $message.'</blockquote>');
		} else {
		    $r->print(' '.&Apache::lonxml::xmlparse($r,'tex',$message).' ');
		}
                 if ($allowed) {
                $r->print('<br /><textarea cols="80" rows="10" name="'.$_.'">'.
			   $syllabus{$_}.
           '</textarea><input type="submit" name="storesyl" value="'.&mt('Store').'" />');
	        }
	    } else {
                if ($target ne 'tex') {
		    $r->print('<h1>'.$message.'</h1>');
		} else {
		    $r->print('\\\\\textbf{'.&Apache::lonxml::xmlparse($r,'tex',$message).'}\\\\');
		}
                if ($allowed) {
                 $r->print(
                '<br />'.&mt('Topic').'<br /><textarea cols="80" rows="2" name="'.$_.'">'.
			   $syllabus{$_}.
           '</textarea><input type="submit" name="storesyl" value="'.&mt('Store').'" />');
                }
            }
	  }
       }
       if ($allowed) {
	   $r->print('</form>');
       }
       if ($target ne 'tex') {$r->print('</p>');} else {$r->print('\\\\');}
    } else {
       $r->print('<p>'.&mt('No page information provided.').'</p>');
    }
    if ($target ne 'tex') { $r->print('</td></tr></table>'); }
    if ($target ne 'tex') {
	$r->print(&Apache::lonfeedback::list_discussion
	           ('board','OPEN','bulletin___'.$marker.'___'.
	            $r->uri));
    } else {
	$r->print('\\\\'.&Apache::lonxml::xmlparse($r,'tex',&Apache::lonfeedback::list_discussion
	           ('board','OPEN','bulletin___'.$marker.'___'.
	            $r->uri)));
    }
    if ($target ne 'tex') {$r->print('</body></html>');} else {$r->print('\end{document}');}
    return OK;
} 

1;
__END__
