# The LearningOnline Network with CAPA
# Handler to show differences between file versions
#
# $Id: londiff.pm,v 1.17 2005/05/29 01:46:16 www Exp $
#
# Copyright Michigan State University Board of Trustees
#
# This file is part of the LearningOnline Network with CAPA (LON-CAPA).
#
# LON-CAPA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LON-CAPA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LON-CAPA; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# /home/httpd/html/adm/gpl.txt
#
# http://www.lon-capa.org/
#
###

package Apache::londiff;

use strict;
use Apache::File;
use File::Copy;
use Algorithm::Diff qw(diff);
use Apache::Constants qw(:common :http :methods);
use Apache::loncacc;
use Apache::lonnet;
use Apache::loncommon();
use Apache::lonlocal;


sub get_split_file {
    my ($fn,$style)=@_;
    my $f1;
    my @f1;
    if ($style='local') {
	if (-e $fn) {
	    my $fh=Apache::File->new($fn);
	    my $line;	
	    while($line=<$fh>) {
		$f1.=$line;
	    }
	}
    } elsif ($style eq 'remote') {
	my $f1=&Apache::lonnet::getfile($fn);
    }
    if ($f1=~/\r/) {
	@f1=split(/\r/,&Apache::lonnet::getfile($fn));      
	foreach my $line (@f1) {
	    $line=~s/\n//g;
	}
    } else {
	@f1=split(/\n/,&Apache::lonnet::getfile($fn));      
    }
    return @f1;
}

sub handler {

  my $r=shift;

# Get query string for limited number of parameters

   &Apache::loncommon::get_unprocessed_cgi($ENV{'QUERY_STRING'},
			   ['filename','versiontwo','versionone','filetwo']);
# Get the files

  my $cuname=$env{'user.name'};
  my $cudom=$env{'user.domain'};

  if ($env{'form.filename'}=~/^\/res\//) {
      ($cudom,$cuname,$env{'form.filename'})=
	  ($env{'form.filename'}=~/^\/res\/(\w+)\/(\w+)\/(.*)$/);
  } else {
      unless (($cuname,$cudom)=
	      &Apache::loncacc::constructaccess($env{'form.filename'},
						$r->dir_config('lonDefDomain'))) {
	  $r->log_reason($cuname.' at '.$cudom.
			 ' trying to get diffs file '.$env{'form.filename'}.
			 '  - not authorized', 
			 $r->filename); 
	  return HTTP_NOT_ACCEPTABLE;
      }
  }
  
  my $efn=$env{'form.filename'};

  $efn=~s/\/\~(\w+)//g;

  my @f1=();
  my @f2=();

  &Apache::loncommon::content_type($r,'text/html');
  $r->send_http_header;

  $r->print('<html><head><title>LON-CAPA Construction Diffs</title></head>');

  $r->print(&Apache::loncommon::bodytag('Resource Differences'));

  
  $r->print('<h1>'.($env{'form.filetwo'}?'':&mt('Compare versions of')).
	    ' <tt>'.$efn.'</tt></h1>');
   
       if (($cuname ne $env{'user.name'}) || ($cudom ne $env{'user.domain'})) {
          $r->print('<h3><font color=red>Co-Author: '.$cuname.' at '.$cudom.
               '</font></h3>');
      }


 if (&Apache::loncommon::fileembstyle(($efn=~/\.(\w+)$/)) eq 'ssi') {
     $r->print('<p><font color="red">');
     if ($env{'form.versionone'} eq 'priv') {
	 my $fn='/home/'.$cuname.'/public_html/'.$efn;
	 @f1=&get_split_file($fn,'local');
	 $r->print('<b>'.&mt('Construction Space Version').'</b>');
     } else {
	 my $fn=
	     '/home/httpd/html//res/'.$cudom.'/'.$cuname.'/';
	 if ($env{'form.versionone'}) {
	     my ($main,$suffix)=($efn=~/^(.+)\.(\w+)$/);
	     $fn.=$main.'.'.$env{'form.versionone'}.'.'.$suffix;
	     $r->print('<b>'.&mt('Version').' '.$env{'form.versionone'}.'</b>');
	 } else {
	     $fn.=$efn;
	     $r->print('<b>'.&mt('Current Version').'</b>');
	 }
	 @f1=&get_split_file($fn,'remote');
     }

     $r->print('</font><br />'.&mt('versus').'<br /><font color="green">');

     if ($env{'form.filetwo'}) {
         my $efn2=$env{'form.filetwo'};
	 $efn2=~s/\/\~(\w+)//g;
	 my $fn='/home/'.$cuname.'/public_html/'.$efn2;
	 @f2=&get_split_file($fn,'local');
	 $r->print('<tt>'.$efn2.'</tt>');
     } elsif ($env{'form.versiontwo'} eq 'priv') {
	 my $fn='/home/'.$cuname.'/public_html/'.$efn;
	 @f2=&get_split_file($fn,'local');
	 $r->print('<b>'.&mt('Construction Space Version').'</b>');
     } else {
	 my $fn=
	     '/home/httpd/html/res/'.$cudom.'/'.$cuname.'/';
	 if ($env{'form.versiontwo'}) {
	     my ($main,$suffix)=($efn=~/^(.+)\.(\w+)$/);
	     $fn.=$main.'.'.$env{'form.versiontwo'}.'.'.$suffix;
	     $r->print('<b>'.&mt('Version').' '.$env{'form.versiontwo'}.'</b>');
	 } else {
	     $fn.=$efn;
	     $r->print('<b>'.&mt('Current Version').'</b>');
	 }
	 @f2=&get_split_file($fn,'remote');
     }
     $r->print('</font></p>');
# Run diff

     my $diffs = diff(\@f1, \@f2);

# Start page output

     my $chunk;
     my $line;

     $r->print('<pre>');

     foreach $chunk (@$diffs) {
	 
	 foreach $line (@$chunk) {
	     my ($sign, $lineno, $text) = @$line;
	     $text=~s/\</\&lt\;/g;
	     $text=~s/\>/\&gt\;/g;
	     $lineno=substr($lineno.'        ',0,7);
	     $r->print('<font color='.(($sign eq '+')?'green':'red').'>'.
		       $sign.' '.$lineno.' '.$text."</font>\n");
	 }
	 $r->print("<hr>\n");
     }
     $r->print('</pre>');
     
 } else {
     $r->print('<h1><font color=red>'.&mt('Binary File').'</font></h1>');
 }
  $r->print('<center><a href="javascript:window.close();">'.&mt('Close This Window').'</a></center></body></html>'); 
  return OK;  
}


1;
__END__




